
"""
DL 通用功能 
"""
import os 

from tpf.box.fil import log as log2 
from tpf import pkl_load,pkl_save
from tpf.d2 import mean_by_count
import matplotlib.pyplot as plt 
from PIL import Image

# PyTorch 三组件
import torch
from torch import nn
from torch.nn import functional as F

# 数据打包工具
from torch.utils.data import Dataset
from torch.utils.data import DataLoader

# 图像处理工具
from matplotlib import pyplot as plt

# import cv2
from torchvision import transforms

# 通用科学计算器
import numpy as np

# 解决 OMP问题
import os,time 
os.environ["KMP_DUPLICATE_LIB_OK"]="TRUE"


BASE_DIR=os.path.dirname(os.path.abspath(__file__))
log_file = "main.log"
log_file = os.path.join(BASE_DIR,log_file)
 
    
def acc2(pre=[],real=[],acc=0.999):
    """
    连续浮点型标签，精度计算
    - 定下一个区间，在区间内为真

    >>> 0.1914*(1-0.95)
        0.00957
    """
    # print("基准：",acc)
    pre = pre.flatten()
    a1 = real.flatten() 
    aa = pre - a1
    
    a2 = np.abs(aa)/(a1+0.00001)  # 防分母为0
    acc_list =np.array([])
    for v in a2:
        if v.item() <= (1-acc):  # 误差线
            acc_list=np.append(acc_list,1)
        else:
            acc_list=np.append(acc_list,0)
    acc3 = acc_list.sum()*100/len(acc_list)
    acc3 = "%.2f"%(acc3) # 前面乘100，后又保留两位有效数据，共四位有效数字
    return acc3 

def deviation2(pre=[],real=[]):
    """
    连续浮点型标签，偏差计算：|real - pred|/|real|

    >>> 0.1914*(1-0.95)
        0.00957
    """

    a1 = real.flatten() 

    index = 0 
    acc_list =np.array([])
    for v in a1:
        aa = np.abs(v-pre[index])
        if np.allclose(v,0):  # 如果真实为0,allclose可计算浮点数是否为0
            acc_list=np.append(acc_list,aa)
        else:
            a2 = aa/v 
            acc_list=np.append(acc_list,a2)
        index += 1 
    acc3 = acc_list.sum()*100/len(acc_list)
    acc3 = "%.6f"%(acc3) # 前面乘100，后又保留两位有效数据，共四位有效数字
    return acc3 

def log(msg, fil="./main.log", max_file_size=10*1024*1024):
    """最大日志，超10M重写日志文件 
    """
    if fil:
        log_file = fil
    else:
        log_file = "/tmp/train.log"
    
    if os.path.exists(log_file):
        fil_size = os.path.getsize(log_file)
        if fil_size > max_file_size:
            # 写入空文件
            with open(log_file,"w",encoding="utf-8") as f:
                tim = time.strftime('%Y-%m-%d %H:%M:%S',time.localtime(time.time()))
                f.write("[{}] {} \n".format(tim,msg))
        else:
            with open(log_file,"a+",encoding="utf-8") as f:
                tim = time.strftime('%Y-%m-%d %H:%M:%S',time.localtime(time.time()))
                f.write("[{}] {} \n".format(tim,msg))
    else:
        with open(log_file,"w",encoding="utf-8") as f:
            tim = time.strftime('%Y-%m-%d %H:%M:%S',time.localtime(time.time()))
            f.write("[{}] {} \n".format(tim,msg))


class T(object):
    """训练器
    """
    is_regression = False
    loss_count = 0
    loss_list = np.array([])

    #深度学习按批次训练，完成一个批次加1
    batch_count = 0 

    #一个轮次遍历一次全体数据集
    epoch = 0

    device = "cuda:0" if torch.cuda.is_available() else "cpu"

    img_save_path = "ml_acc.jpg"
    acc_test_file  ="ml_acc_test.txt"
    acc_train_file ="ml_acc_train.txt"
    loss_item_file="ml_loss_item.txt"
    good_model_path="ml_good_model_acc.txt"
    good_num=0
    optim_save_path="ml_optim_params.pkl"
    epoch_num_path="ml_epoch_num.txt"
    log_input_pred="ml_input_pre.log"
    log_input_eval="ml_input_eval.log"
    log_dataset = "ml_dataset.log"


 
    def __init__(self, os_type="linux") -> None:
        """
        训练模板类
        """
        # if os_type == "linux":
        # 设备
        # self.device = "cuda:0" if torch.cuda.is_available() else "cpu"
    
        pass 

    # 定义过程监控函数
    @classmethod
    def _get_acc(cls, dataloader, model):
        """精度计算
        """
        accs = []
        model.eval()
        with torch.no_grad():
            for X4,y4 in dataloader:
                y_pred = model(X4)
                log(msg=f"\n eval input x0:\n{X4},\ny0:\n{y4[0]},\nmodel pre:\n{y_pred[0]}",fil=cls.log_input_eval)
                if cls.is_regression:
                    acc = acc2(pre=y_pred,real=y4)
                    return acc
                else:
                    y_pred = y_pred.argmax(dim=1)
                    label1_mean = y_pred[y4==1].float().mean()
                    
                    msg = f"\n batch size:{y4.shape[0]},pre right:{(y_pred == y4).float().mean()},pre1:{label1_mean}"
                    # print(msg)
                    log(msg=msg, fil=cls.log_input_eval)

                    acc = (y_pred == y4).float().mean().item()
                    accs.append(acc)
        return np.array(accs).mean()

    @classmethod
    def show_acc(cls, model, train_dataset, test_dataset=None, batch_size=128):
        """过程监控
        """
        train_dataloader = DataLoader(dataset=train_dataset, shuffle=False, batch_size=batch_size)
        if test_dataset:
            test_dataloader = DataLoader(dataset=test_dataset, shuffle=False, batch_size=batch_size)
            acc_train = cls._get_acc(model=model, dataloader=train_dataloader)
            acc_test = cls._get_acc(model=model, dataloader=test_dataloader)
            msg = f"train_acc: {acc_train}, test_acc: {acc_test}, loss:{cls.loss_avg100()}"
            print(msg)
            log(msg=msg,fil=cls.log_file)
            cls.msg_record(val=acc_train,file_path=cls.acc_train_file)
            cls.msg_record(val=acc_test,file_path=cls.acc_test_file)
            return acc_train,acc_test
        else:
            acc_train = cls._get_acc(model=model, dataloader=train_dataloader)
            msg = f"train_acc: {acc_train}, loss:{cls.loss_avg100()}"
            print(msg)
            log(msg=msg,fil=cls.log_file)
            cls.msg_record(val=acc_train,file_path=cls.acc_train_file)
            return acc_train,None


    @classmethod
    def diff_time(cls, t=None, return_min=False):
        """代码运行时间差判断
        """
        import time 
        b = time.time()
        if not t:
            return b
        
        if return_min:
            c = round((b-t)/60)
            print(f"time pass {c} min")
        else:
            c = round(b-t)
            print(f"time pass {c} sec")

        return b


    @classmethod
    def loss_item_compute(cls, loss):
        """循环计算最近100个批次损失值"""

        # 当前批次损失
        cls._loss_item = loss 

        cls.loss_count += 1

        if cls.loss_count <= 100:
            cls.loss_list = np.append(cls.loss_list, loss)
        else:
            i = cls.loss_count%100
            cls.loss_list[i] = loss

        cls.msg_record(val=loss,file_path=cls.loss_item_file)

    @classmethod
    def val_list(cls, file_path):
        vals = []
        with open(file_path) as f:
            line = f.readline() 
            while line:
                line = line.strip()
                if len(line)>0:
                    vals.append(float(line))
                line = f.readline()
        return vals

    @classmethod
    def acc_test_list(cls):
        return cls.val_list(cls.acc_test_file)

    @classmethod
    def acc_train_list(cls):
        return cls.val_list(cls.acc_train_file)

    @classmethod
    def loss_item_list(cls):
        return cls.val_list(cls.loss_item_file)

    @classmethod
    def show_img_test(cls):
        vals = cls.acc_test_list()
        plt.plot(vals,color='r',linewidth=0.8)
    
    @classmethod
    def show_img_train(cls):
        vals = cls.acc_train_list()
        plt.plot(vals,color='g',linewidth=0.8)

    @classmethod
    def show_img(cls, using_old=False, mean_num=1):

        if os.path.exists(cls.img_save_path) and using_old:
            
            img = Image.open(cls.img_save_path)
            return img
        else:
            axis_y1 = cls.acc_train_list()
            axis_y2 = cls.acc_test_list()

            axis_y1 = mean_by_count(axis_y1,mean_num=mean_num)
            axis_y2 = mean_by_count(axis_y2,mean_num=mean_num)


            axis_x = [i for i in range(len(axis_y2))]

            # 显示设置
            plt.rcParams['axes.unicode_minus'] = False   # 负号显示
            plt.rcParams['font.style'] = 'normal'  # 正体normal， 斜体italic
            plt.rcParams['font.size'] = 14  # 字号

            fig, ax = plt.subplots(figsize=(8, 6))
            ax.plot(axis_x, axis_y1, label="train", color='g',linewidth=0.8)
            ax.plot(axis_x, axis_y2, label="test", color='r',linewidth=0.8,alpha=0.6)

            ax.set_xlabel("epoch")  # x轴标签
            ax.set_ylabel("p")  # y轴标签
            ax.set_title("acc")     # 标题
            plt.grid(visible=True, linestyle="--", alpha=0.5)  # 显示栅格
            plt.legend()  # 显示图例
            plt.savefig(cls.img_save_path, bbox_inches='tight', pad_inches=0.2)  # 保存
            plt.show()
            plt.close()
            

    @classmethod
    def show_img_loss(cls,mean_num=1, s=4,  marker='o',alpha=0.9,linewidths=0.3):
        vals = cls.loss_item_list()
        vals = mean_by_count(vals,mean_num=mean_num)

        axis_x = np.array([i for i in range(len(vals))])
        axis_y = np.array(vals)
        #根据x,y生成一个数列，将对应的数据点映射到coclarmap中的颜色上
        C = np.arctan(axis_x/axis_y)

        #c=C,edgecolors='w'
        plt.scatter(axis_x, axis_y, s=s,  marker=marker,alpha=alpha,linewidths=linewidths)


    @classmethod
    def write(cls,obj,file_path):
        """
        直接将对象转字符串写入文件,这样可以在文件打开时,看到原内容,还可以进行搜索
        """
        ss = str(obj)
        with open(file_path,"w",encoding="utf-8") as f:
            f.write(ss)

    @classmethod
    def read(cls,file_path):
        with open(file_path,'r',encoding="utf-8") as f:
            c = eval(f.read())
            return c 


    @classmethod
    def get_good_model_acc(cls):
        if os.path.exists(cls.good_model_path):
            return cls.read(cls.good_model_path)
        else:
            return 0
            

    @classmethod
    def set_good_model_acc(cls,acc):
        cls.write(acc, file_path=cls.good_model_path)


    @classmethod
    def get_epoch_num(cls):
        if os.path.exists(cls.epoch_num_path):
            return cls.read(cls.epoch_num_path)
        else:
            return 0
            
    @classmethod
    def set_epoch_num(cls):
        epoch_num = cls.get_epoch_num()+1
        cls.write(epoch_num, file_path=cls.epoch_num_path)


    @classmethod
    def msg_record(cls, val, file_path="acc_test.txt",replace=False):
        """记录每个批次的信息于文件中，一行一个记录"""

        if cls.epoch == 1 and replace:
            # print("cls.batch_count:",cls.batch_count,file_path)
            with open(file_path, "w", encoding="utf-8") as f:
                f.write(f"{round(val, ndigits=3)}\n")
        else:
            with open(file_path, "a+", encoding="utf-8") as f:
                f.write(f"{round(val, ndigits=3)}\n")
            
            
        
    @classmethod
    def loss_item(cls):
        return cls._loss_item

    @classmethod
    def loss_avg100(cls):
        item = round(cls.loss_list.mean(), ndigits=4)
        return item

    @classmethod
    def train_one_epoch(cls,
            model,
            loss_fn, 
            optimizer, 
            train_dataloader=None):

        count = 0 

        model.to(device=cls.device)
        model.train()

        for X3,y3 in train_dataloader:
            X3.to(device=cls.device)
            y3.to(device=cls.device)

            count += 1
            cls.batch_count += 1
            

            # 前一段时间看看小的批次，后面开始训练后，就不看了
            if count %10 ==0 and cls.batch_count < 1000: 
                print(f"batch {count} start...")

            # 正向传播
            y_pred = model(X3)

            # 清空梯度
            optimizer.zero_grad()

            # 计算损失
            loss = loss_fn(y_pred, y3)

            # 梯度下降
            loss.backward()

            # 优化一步
            optimizer.step()

            with torch.no_grad():
                cls.loss_item_compute(loss.item())

                # 输出
                log(msg=f"\ntrain input x0:\n{X3},\ny0:\n{y3[0]},\nmodel pre:\n{y_pred[0]}",fil=cls.log_input_pred)
                
                if loss < 0.01 and cls.loss_count%100==99:
                    log(msg=f"y_pred[:3]:{y_pred[:3]}",fil=cls.log_file)
                    log(msg=f"y_real[:3]:{y3[:3]}",fil=cls.log_file)

                # 前一段时间看看小的批次，后面开始训练后，就不看了
                if count %30 ==0 : 
                    # if cls.continuation:
                    #     torch.save(model.state_dict(), cls.param_path)
                    msg = f"batch {count}, loss:{loss.item()}"
                    print(msg)
                    log(msg=msg, fil=cls.log_file)
                

            
        return loss


    # 定义训练过程
    @classmethod
    def train(cls,
            model, 
            loss_fn,
            optimizer="sgd", 
            train_dataset_func=None, test_dataset_func=None,
            train_dataset=None, test_dataset=None,
            train_dataloader=None,
            epochs=1, 
            batch_size=128,
            learning_rate=1e-3,
            model_param_path="ml_model1_params.h5",
            auto_save=True,
            continuation=True,
            is_regression=False,
            log_file="/tmp/train.log"):
        """
        梯度下降训练
        --------------------------
        optimizer:"adam","sgd", 
        train_dataloader: 指定此值，全局使用这一个loader，否则每次开始，重新生成一个loader

        auto_save:是否自动保存模型参数，

        continuation:是否持续训练，

        model_param_path:模型参数保存路径
        is_regression: True-回归问题，False-分类问题

        """
        param_path= model_param_path
        cls.param_path = param_path
        cls.is_regression= is_regression
        cls.log_file = log_file
        cls.continuation = continuation

        # 加载上次训练的参数
        if continuation:
            if os.path.exists(param_path):
                print(f"load param:{param_path}")
                model.load_state_dict(torch.load(param_path, map_location=cls.device))
                
                if os.path.exists(cls.optim_save_path):
                    optimizer = pkl_load(file_path=cls.optim_save_path, use_joblib=False)
                
            else:
                print("重新开始训练...")
                
        # 定义优化器,重新加载的优化器数据类型不会为str
        if isinstance(optimizer,str):
            if optimizer.lower() == "sgd":
                optimizer = torch.optim.SGD(params=model.parameters(), lr=learning_rate)
            elif optimizer.lower() == "adam":
                optimizer = torch.optim.Adam(params=model.parameters(), lr=learning_rate)

        

        t_start = cls.diff_time()
        t = t_start
        for epoch in range(1, epochs+1):
            cls.epoch = epoch
            if train_dataset_func:
                train_dataset = train_dataset_func()

            if test_dataset_func:
                test_dataset = test_dataset_func()

            msg = f"\ntrain_dataset[0]:\n {train_dataset[0]},\ntest_dataset[0]:{test_dataset[0]}\n"
            log(msg, fil=cls.log_dataset)
            
            if not train_dataloader:
                # 在每轮训练开始时，重新启用一个loader，理论上这种才合理，实际上，这种方法精度比loader全局定义的方法，每一轮精度低一些
                train_dataloader = DataLoader(dataset=train_dataset, shuffle=True, batch_size=batch_size)
            
            all_batch_size = len(train_dataloader)
            msg = f"\n正在进行第 epoch= {epoch} 轮训练...每轮次批次个数 ={all_batch_size}"
            print(msg)
            log(msg, fil=log_file)
            T.train_one_epoch(model=model, loss_fn=loss_fn,optimizer=optimizer,train_dataloader=train_dataloader)
                
            acc_train,acc_test = T.show_acc(model=model, train_dataset=train_dataset, test_dataset=test_dataset)
            # print(f"train_acc: {show_acc(dataloader=train_dataloader)}, test_acc: {show_acc(dataloader=test_dataloader)}")

            
            # 保存模型（参数）
            # 保留最好的三个模型以及当前的模型
            if continuation or auto_save:
                torch.save(model.state_dict(), cls.param_path)
                pkl_save(optimizer, file_path=cls.optim_save_path, use_joblib=False)
                # print(f"save param:{param_path}")
                if acc_test:
                    good_acc = cls.get_good_model_acc()
                    if acc_test > good_acc or cls.epoch == 1:
                        good_params_path = "ml_{}_{}.good".format(cls.param_path, cls.good_num%3)
                        torch.save(model.state_dict(), good_params_path)
                        cls.set_good_model_acc(acc_test)


            t = cls.diff_time(t)

"""
T.train(
    continuation=True,
    model=lenet,loss_fn=loss_fn,optimizer=optimizer,
    epochs=30,
    train_dataset=train_dataset,test_dataset=test_dataset,train_dataloader=train_dataloader)

"""

class Activate(object):
    
    def __init__(self):
        pass 

    @staticmethod
    def relu(X):
        X = np.array(X)
        shape = X.shape
        X = X.reshape(-1)
        x = np.array([0 if x < 0 else x for x in X])
        x = x.reshape(shape)
        return x
    
    @staticmethod
    def relu2(X):
        X = torch.tensor(data=X,dtype=torch.float32)
        return X.relu()
    
    @staticmethod
    def relu3(X):
        X = torch.tensor(data=X,dtype=torch.float32)
        return F.relu(input=X,inplace=True)
    
    @staticmethod
    def rule_test():
        
        np.random.seed(73)
        A=np.random.randn(7,3)
        # print(A)
        """
        [[ 0.57681305  2.1311088   2.44021967]
        [ 0.26332687 -1.49612065 -0.03673531]
        [ 0.43069579 -1.52947433 -0.73025968]
        [ 1.05131524  1.61979267 -1.60501337]
        [ 0.33100953 -0.21095236  0.2981767 ]
        [-1.14607352  0.57536202 -0.36390663]
        [ 0.03639919 -0.52056399 -0.01576433]]
        """

        # B = Activate.relu(X=A)
        # print(B)
        """
        [[0.57681305 2.1311088  2.44021967]
        [0.26332687 0.         0.        ]
        [0.43069579 0.         0.        ]
        [1.05131524 1.61979267 0.        ]
        [0.33100953 0.         0.2981767 ]
        [0.         0.57536202 0.        ]
        [0.03639919 0.         0.        ]]
        """

        # B = Activate.relu2(X=A)
        # print(B)
        """可以看出torch.relu()只保留了四位有效数字
        tensor([[0.5768, 2.1311, 2.4402],
                [0.2633, 0.0000, 0.0000],
                [0.4307, 0.0000, 0.0000],
                [1.0513, 1.6198, 0.0000],
                [0.3310, 0.0000, 0.2982],
                [0.0000, 0.5754, 0.0000],
                [0.0364, 0.0000, 0.0000]], dtype=torch.float64)
        """

        B = Activate.relu3(X=A)
        print(B)
        """F.relu同样改变了数据的精度
        tensor([[0.5768, 2.1311, 2.4402],
                [0.2633, 0.0000, 0.0000],
                [0.4307, 0.0000, 0.0000],
                [1.0513, 1.6198, 0.0000],
                [0.3310, 0.0000, 0.2982],
                [0.0000, 0.5754, 0.0000],
                [0.0364, 0.0000, 0.0000]], dtype=torch.float64)
        """


